// Copyright © 2024 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#include <fstream>
#include <sstream>
#include <vector>
#include <iterator>
#include <cstring>
#include <filesystem>
#include <pthread.h>
#include <algorithm>
#include <numeric>
#include <random>
#include <functional>
#include <sys/stat.h>
#include <sys/mman.h>
#include <math.h>
#include <thread>
#include <chrono>
#include <limits>
#include <queue>
#include <string>
#include <cctype>
#include <unordered_set>
#include "es_qwen2_client.hpp"
#include "es_npu_interface.h"
#include "es_sys_memory.h"
#include "json.h"
#include "half.h"

namespace fs = std::filesystem;

static int32_t client_log_level = 0;

#define client_log(func, line, level, fmt, ...)                                                                \
    do {                                                                                                       \
        if (level <= client_log_level)                                                                         \
            printf("[ES][Qwen2][%s](%s:%d) " fmt, level == 0 ? "ERROR" : "DEBUG", func, line, ##__VA_ARGS__); \
    } while (0)

#define ES_LOG_DBG(format, ...) client_log(__func__, __LINE__, 2, format, ##__VA_ARGS__)
#define ES_LOG_MSG(format, ...) client_log(__func__, __LINE__, 1, format, ##__VA_ARGS__)
#define ES_LOG_ERR(format, ...) client_log(__func__, __LINE__, 0, format, ##__VA_ARGS__)

template <typename T>
void check(T result, char const* const func, ES_S32 const line)
{
    if (result) {
        fprintf(stderr, "%s(%d) error=%d \n", func, line, static_cast<unsigned ES_S32>(result));
        exit(EXIT_FAILURE);
    }
}

#define esCheckError(val) check((val), __func__, __LINE__)

#define DUMP_TOKEN_PREFIX(maskVal, layer) \
    ("token_" + std::to_string(maskVal) + "_" + "layer_" + std::to_string(layer))

Qwen2Client::Qwen2Client(const std::string& configPath)
    : mWordEmbeddings(nullptr)
{
    ES_LOG_DBG("start to load %s\n", configPath.c_str());
    if (loadConfig(configPath)) {
        throw std::runtime_error("Config file is not exist.");
    }

    mTokenizer = std::make_unique<QwenTokenizer>(mConfig.tiktokenFile);
    if (!mTokenizer) {
        throw std::runtime_error("Construct QwenTokenizer failed.");
    }

    if (loadEmbedding()) {
        throw std::runtime_error("Load embedding failed.");
    }

    ES_LOG_DBG("start to load model from directory.\n");
    if (loadModel()) {
        throw std::runtime_error("Load model failed.");
    }

    ES_LOG_DBG("Qwen2 Client init done.\n");
}

Qwen2Client::~Qwen2Client()
{
    uint16_t deviceId = 0;

    for (auto& blockTask : mBlockTasks) {
        esCheckError(ES_NPU_UnloadModel(blockTask.task.modelId));
    }

    if (mConfig.logits) {
        esCheckError(ES_NPU_UnloadModel(mLogitsTask.task.modelId));
    }

    releaseTaskBuffers();

    delete[] mWordEmbeddings;

    esCheckError(ES_NPU_ReleaseDevice(deviceId));
}

ES_S32 Qwen2Client::releaseTaskBuffers()
{
    MODEL_TASK_S *blockTask = &mBlockTasks.front();
    esCheckError(ES_NPU_Unprepare(&blockTask->task.inputFd[IN_TOKEN_IDX]));
    ES_SYS_Munmap(blockTask->inputMapAddr[IN_TOKEN_IDX], blockTask->task.inputFd[IN_TOKEN_IDX].size);
    ES_SYS_MemFree(blockTask->task.inputFd[IN_TOKEN_IDX].memFd);

    esCheckError(ES_NPU_Unprepare(&blockTask->task.outputFd[OUT_HIDDEN_IDX]));
    ES_SYS_Munmap(&blockTask->outputMapAddr[OUT_HIDDEN_IDX], blockTask->task.outputFd[OUT_HIDDEN_IDX].size);
    ES_SYS_MemFree(blockTask->task.outputFd[OUT_HIDDEN_IDX].memFd);

    for (auto& blockTask : mBlockTasks) {
        esCheckError(ES_NPU_Unprepare(&blockTask.task.inputFd[IN_K_CACHE_IDX]));
        ES_SYS_Munmap(blockTask.inputMapAddr[IN_K_CACHE_IDX], blockTask.task.inputFd[IN_K_CACHE_IDX].size);
        ES_SYS_MemFree(blockTask.task.inputFd[IN_K_CACHE_IDX].memFd);

        esCheckError(ES_NPU_Unprepare(&blockTask.task.inputFd[IN_V_CACHE_IDX]));
        ES_SYS_Munmap(blockTask.inputMapAddr[IN_V_CACHE_IDX], blockTask.task.inputFd[IN_V_CACHE_IDX].size);
        ES_SYS_MemFree(blockTask.task.inputFd[IN_V_CACHE_IDX].memFd);

        esCheckError(ES_NPU_Unprepare(&blockTask.task.inputFd[IN_MASK_IDX]));
        ES_SYS_Munmap(blockTask.inputMapAddr[IN_MASK_IDX], blockTask.task.inputFd[IN_MASK_IDX].size);
        ES_SYS_MemFree(blockTask.task.inputFd[IN_MASK_IDX].memFd);
    }

    if (mConfig.logits) {
        esCheckError(ES_NPU_Unprepare(&mLogitsTask.task.inputFd[0]));
        ES_SYS_Munmap(mLogitsTask.inputMapAddr[0], mLogitsTask.task.inputFd[0].size);
        ES_SYS_MemFree(mLogitsTask.task.inputFd[0].memFd);

        esCheckError(ES_NPU_Unprepare(&mLogitsTask.task.outputFd[0]));
        ES_SYS_Munmap(mLogitsTask.outputMapAddr[0], mLogitsTask.task.outputFd[0].size);
        ES_SYS_MemFree(mLogitsTask.task.outputFd[0].memFd);
    }

    return ES_SUCCESS;
}

void Qwen2Client::printFp16Vec(const std::vector<fp16_t>& vec, int32_t num)
{
    float tmp;
    for (int32_t idx = 0; idx < num; idx++) {
        tmp = half_float::detail::half2float(vec[idx]);
        ES_LOG_DBG("%f\n", tmp);
    }
}

ES_S32 Qwen2Client::dumpData(const void* address, const size_t size, const std::string& file)
{
    std::string outputPath = std::string("dump/") + file;
    std::ofstream outputFile(outputPath, std::ios::binary);
    if (!outputFile.is_open()) {
        ES_LOG_ERR("Failed to open file: %s\n", outputPath.c_str());
        return ES_NPU_ERROR_MODEL_NOT_FOUND;
    }

    outputFile.write((const char*)address, size);
    outputFile.close();

    ES_LOG_DBG("\n===> Data dumped to file: %s\n", outputPath.c_str());
    return ES_SUCCESS;
}

ES_S32 Qwen2Client::loadConfig(const std::string& configPath)
{
    std::ifstream configFile(configPath, std::ifstream::binary);
    if (!configFile.is_open()) {
        return ES_NPU_ERROR_MODEL_NOT_FOUND;
    }

    Json::Value root;
    configFile >> root;

    const Json::Value modelDir = root["model_dir"];
    mConfig.modelDir = modelDir.asString();

    const Json::Value blockModelList = root["block_models"];
    for (const auto& model : blockModelList) {
        mConfig.blockModelFiles.push_back(model.asString());
    }

    if (mConfig.blockModelFiles.size() == 0) {
        ES_LOG_ERR("There is no block model in %s\n", mConfig.modelDir.c_str());
        return ES_NPU_ERROR_MODEL_NOT_FOUND;
    }

    const Json::Value logits = root["logits"];
    mConfig.logits = logits.asBool();
    if (mConfig.logits) {
        const Json::Value logitsModel = root["logits_models"];
        mConfig.logitsModelFile = logitsModel.asString();
    }

    mConfig.tiktokenFile = root["tokenizer_file"].asString();
    mConfig.embeddingFile = root["embedding_file"].asString();
    mConfig.vocabSize = root["vocab_size"].asInt();
    mConfig.numHead = root["num_head"].asInt();
    mConfig.embeddingSize = root["embedding_size"].asInt();
    mConfig.embeddingDim = root["embedding_dim"].asInt();
    mConfig.embeddingDim1 = root["embedding_dim1"].asInt();
    mConfig.embeddingDim2 = root["embedding_dim2"].asInt();
    mConfig.tokenNum = root["token_num"].asInt();
    mConfig.tokenN1 = root["token_n1"].asInt();
    mConfig.tokenN2 = root["token_n2"].asInt();
    mConfig.precision = root["precision"].asInt();

    mConfig.language = root["language"].asString();

    const Json::Value eosList = root["eos_list"];
    for (const auto& eos : eosList) {
        mConfig.eosList.push_back(eos.asString());
    }

    mConfig.responseMaxLen = root["response_max_len"].asInt();
    mConfig.responseLineLen = root["response_line_len"].asInt();
    mConfig.logLevel = root["log_level"].asInt();
    mConfig.randomCnt = root["random_cnt"].asInt();
    mConfig.repeatLen = root["repeat_len"].asInt();
    mConfig.repeatCnt = root["repeat_cnt"].asInt();
    mConfig.dataDump = root["data_dump"].asInt();
    mConfig.dataTokenInterval.min = root["data_dump_token"][0].asInt();
    mConfig.dataTokenInterval.max = root["data_dump_token"][1].asInt();
    mConfig.perfEnable = root["perf_enable"].asInt();
    mConfig.opDump = root["op_dump"].asInt();
    mConfig.opType = root["op_dump_type"].asInt();
    mConfig.opTokenInterval.min = root["op_dump_token"][0].asInt();
    mConfig.opTokenInterval.max = root["op_dump_token"][1].asInt();
    mConfig.opLayerInterval.min = root["op_dump_layer"][0].asInt();
    mConfig.opLayerInterval.max = root["op_dump_layer"][1].asInt();
    mConfig.opListInterval.min = root["op_dump_list"][0].asInt();
    mConfig.opListInterval.max = root["op_dump_list"][1].asInt();
    mConfig.modelTest = root["model_test"].asInt();
    mConfig.repetitionPenalty = root["repetition_penalty"].asFloat();
    mConfig.topK = root["top_k"].asInt();

    const Json::Value& prompts = root["prompts"];
    for (Json::ValueConstIterator it = prompts.begin(); it != prompts.end(); ++it) {
        int32_t key = std::stoi(it.key().asString());
        mConfig.prompts[key] = it->asString();
    }

    client_log_level = mConfig.logLevel;

    return ES_SUCCESS;
}

ES_S32 Qwen2Client::loadEmbedding()
{
    try {
        mWordEmbeddings = new fp16_t[mConfig.vocabSize * mConfig.embeddingSize];
    } catch (const std::bad_alloc& e) {
        ES_LOG_ERR("alloc memory for word embedding failed\n");
        return ES_FAILURE;
    }

    std::ifstream file(mConfig.embeddingFile, std::ios::binary | std::ios::ate);
    if (!file.is_open()) {
        ES_LOG_ERR("Failed to open embedding file: %s\n",mConfig.embeddingFile.c_str());
        return ES_FAILURE;
    }

    std::streampos fileSize = file.tellg();
    if (static_cast<int>(fileSize) != mConfig.vocabSize * mConfig.embeddingSize * sizeof(fp16_t)) {
        ES_LOG_ERR("Invalid embedding file size: %d\n", static_cast<int>(fileSize));
        return ES_FAILURE;
    }

    file.seekg(0, std::ios::beg);
    file.read(reinterpret_cast<char*>(mWordEmbeddings), fileSize);

    if (!file) {
        ES_LOG_ERR("Error reading embedding file.\n");
        return ES_FAILURE;
    }

    file.close();

    return ES_SUCCESS;
}

void Qwen2Client::displayLoadModelProgress(int32_t current, int32_t total)
{
    float progress = (float)current / total * 100.0f;

    std::cout << "\rLoading models: [";
    int pos = 50 * progress / 100.0;
    for (int i = 0; i < 50; ++i) {
        if (i < pos)
            std::cout << "=";
        else if (i == pos)
            std::cout << ">";
        else
            std::cout << " ";
    }
    std::cout << "] " << std::setw(6) << std::setfill(' ') << std::fixed << std::setprecision(2) << progress << "%";
    std::cout.flush();
}

ES_S32 Qwen2Client::loadModel()
{
    uint16_t deviceNum = 0;
    uint16_t deviceId = 0;

    ES_LOG_DBG("set log level:%u\n", mConfig.logLevel);
    esCheckError(ES_NPU_SetLogLevel(mConfig.logLevel));

    esCheckError(ES_NPU_GetNumDevices(&deviceNum));
    if (deviceId >= deviceNum) {
        ES_LOG_ERR("deviceId(%d) is not less than deviceNum(%d)\n", deviceId, deviceNum);
        return ES_NPU_ERROR_DEV_NOT_FOUND;
    }

    ES_LOG_DBG("open device:%u\n", deviceId);
    esCheckError(ES_NPU_SetDevice(deviceId));

    int32_t loadedModels = 0;
    int32_t totalModels = mConfig.blockModelFiles.size();
    uint32_t modelId;

    auto start = std::chrono::high_resolution_clock::now();
    // load  block model
    for (int32_t num = 0; num < totalModels; num++) {
        std::string model = mConfig.modelDir + mConfig.blockModelFiles[num];

        esCheckError(ES_NPU_LoadModelFromFile(&modelId, model.c_str()));
        ES_LOG_DBG("load %d model:%s, modelId=%u\n", num, model.c_str(), modelId);

        esCheckError(prepareBlockTaskBuffers(num, modelId));

        loadedModels++;
        displayLoadModelProgress(loadedModels, totalModels);
    }

    // load logits model
    if (mConfig.logits) {
        std::string model = mConfig.modelDir + mConfig.logitsModelFile;
        esCheckError(ES_NPU_LoadModelFromFile(&modelId, model.c_str()));
        ES_LOG_DBG("load logits model:%s, modelId=%u\n", model.c_str(), modelId);

        esCheckError(prepareLogitsTaskBuffers(modelId));
    }

    auto end = std::chrono::high_resolution_clock::now();
    std::chrono::duration<double> elapsed = end - start;
    std::cout << " (" << elapsed.count() << " seconds)" << std::endl;

    return ES_SUCCESS;
}

ES_S32 Qwen2Client::prepareTensorBuffer(NPU_TENSOR_S &tensor, ES_DEV_BUF_S &devBuf, void **pAddr, bool cache)
{
    ES_U64 memFd;
    SYS_CACHE_MODE_E cacheMode = cache ? SYS_CACHE_MODE_CACHED : SYS_CACHE_MODE_NOCACHE;

    esCheckError(ES_SYS_MemAlloc(&memFd, cacheMode, "qwen2_demo", NULL, tensor.bufferSize));
    *pAddr = ES_SYS_Mmap(memFd, tensor.bufferSize, cacheMode);
    if (!*pAddr) {
        ES_LOG_ERR("ES_SYS_Mmap failed!\n");
        return ES_FAILURE;
    }

    (void)memset(*pAddr, 0x00, tensor.bufferSize);
    if (cache) {
        esCheckError(ES_SYS_MemFlushCache(memFd));
    }

    devBuf.size = tensor.bufferSize;
    devBuf.offset = 0;
    devBuf.memFd = memFd;

    return ES_SUCCESS;
}

ES_S32 Qwen2Client::prepareKVCacheBuffers(uint32_t modelId, MODEL_TASK_S &modelTask)
{
    void* pData;
    ES_U64 memFd;
    ES_U64 kCacheSize = mConfig.precision * mConfig.numHead * mConfig.embeddingDim1 * mConfig.tokenNum * mConfig.embeddingDim2;
    ES_U64 vCacheSize = mConfig.precision * mConfig.numHead * mConfig.tokenN1 * mConfig.embeddingDim * mConfig.tokenN2;
    ES_U64 maskSize = MASK_BUFFER_SIZE;
    NPU_TENSOR_S tensor;

    // Prepare K cache buffer
    esCheckError(ES_NPU_GetInputTensorDesc(modelId, IN_K_CACHE_IDX, &tensor));
    if (tensor.bufferSize != kCacheSize) {
        ES_LOG_ERR("k cache size is not valid, %d!\n", tensor.bufferSize);
        return ES_FAILURE;
    }
    esCheckError(prepareTensorBuffer(tensor, modelTask.task.inputFd[IN_K_CACHE_IDX],
                                     &modelTask.inputMapAddr[IN_K_CACHE_IDX], false));

    modelTask.task.outputFd[OUT_K_CACHE_IDX] = modelTask.task.inputFd[IN_K_CACHE_IDX];
    modelTask.outputMapAddr[OUT_K_CACHE_IDX] = modelTask.inputMapAddr[IN_K_CACHE_IDX];
    esCheckError(ES_NPU_Prepare(&modelTask.task.inputFd[IN_K_CACHE_IDX]));

    // Prepare V cache buffer
    esCheckError(ES_NPU_GetInputTensorDesc(modelId, IN_V_CACHE_IDX, &tensor));
    if (tensor.bufferSize != vCacheSize) {
        ES_LOG_ERR("v cache size is not valid, %d!\n", tensor.bufferSize);
        return ES_FAILURE;
    }
    esCheckError(prepareTensorBuffer(tensor, modelTask.task.inputFd[IN_V_CACHE_IDX],
                                     &modelTask.inputMapAddr[IN_V_CACHE_IDX], false));

    modelTask.task.outputFd[OUT_V_CACHE_IDX] = modelTask.task.inputFd[IN_V_CACHE_IDX];
    modelTask.outputMapAddr[OUT_V_CACHE_IDX] = modelTask.inputMapAddr[IN_V_CACHE_IDX];
    esCheckError(ES_NPU_Prepare(&modelTask.task.inputFd[IN_V_CACHE_IDX]));

    ES_LOG_DBG("prepare KVCache buffers done.\n");

    return ES_SUCCESS;
}

ES_S32 Qwen2Client::prepareBlockTaskBuffers(int32_t modelIdx, uint32_t modelId)
{
    MODEL_TASK_S modelTask;
    NPU_TENSOR_S tensor;
    int32_t numInputTensors, numOutputTensors;

    memset(&modelTask, 0, sizeof(MODEL_TASK_S));
    esCheckError(ES_NPU_GetNumInputTensors(modelId, &numInputTensors));
    if (numInputTensors != IN_MAX_IDX) {
        ES_LOG_ERR("Input tensor num is invalid from model, numInputTensors=%d\n", numInputTensors);
        return ES_NPU_ERROR_INVALID_SIZE;
    }
    modelTask.task.inputFdNum = numInputTensors;

    esCheckError(ES_NPU_GetNumOutputTensors(modelId, &numOutputTensors));
    if (numOutputTensors != OUT_MAX_IDX) {
        ES_LOG_ERR("Output tensor num is invalid from model, numOutputTensors=%d\n", numOutputTensors);
        return ES_NPU_ERROR_INVALID_SIZE;
    }
    modelTask.task.outputFdNum = numOutputTensors;

    // prepare KV Cache
    esCheckError(prepareKVCacheBuffers(modelId, modelTask));

    // preapre mask buffer
    esCheckError(ES_NPU_GetInputTensorDesc(modelId, IN_MASK_IDX, &tensor));
    esCheckError(prepareTensorBuffer(tensor, modelTask.task.inputFd[IN_MASK_IDX],
                                     &modelTask.inputMapAddr[IN_MASK_IDX], false));
    esCheckError(ES_NPU_Prepare(&modelTask.task.inputFd[IN_MASK_IDX]));

    // prepare token buffer
    if (modelIdx == 0) {
        esCheckError(ES_NPU_GetInputTensorDesc(modelId, IN_TOKEN_IDX, &tensor));
        esCheckError(prepareTensorBuffer(tensor, modelTask.task.inputFd[IN_TOKEN_IDX],
                                         &modelTask.inputMapAddr[IN_TOKEN_IDX], false));
        esCheckError(ES_NPU_Prepare(&modelTask.task.inputFd[IN_TOKEN_IDX]));

        esCheckError(ES_NPU_GetOutputTensorDesc(modelId, OUT_HIDDEN_IDX, &tensor));
        esCheckError(prepareTensorBuffer(tensor, modelTask.task.outputFd[OUT_HIDDEN_IDX],
                                         &modelTask.outputMapAddr[OUT_HIDDEN_IDX], false));
        esCheckError(ES_NPU_Prepare(&modelTask.task.outputFd[OUT_HIDDEN_IDX]));
    } else {
        modelTask.task.inputFd[IN_TOKEN_IDX] = mBlockTasks[modelIdx - 1].task.outputFd[OUT_HIDDEN_IDX];
        modelTask.inputMapAddr[IN_TOKEN_IDX] = mBlockTasks[modelIdx - 1].outputMapAddr[OUT_HIDDEN_IDX];

        modelTask.task.outputFd[OUT_HIDDEN_IDX] = mBlockTasks[modelIdx - 1].task.inputFd[IN_TOKEN_IDX];
        modelTask.outputMapAddr[OUT_HIDDEN_IDX] = mBlockTasks[modelIdx - 1].inputMapAddr[IN_TOKEN_IDX];
    }

    modelTask.task.modelId = modelId;
    mBlockTasks.emplace_back(modelTask);

    return ES_SUCCESS;
}

ES_S32 Qwen2Client::prepareLogitsTaskBuffers(uint32_t modelId)
{
    NPU_TENSOR_S tensor;
    int32_t numInputTensors, numOutputTensors;

    memset(&mLogitsTask, 0x00, sizeof(MODEL_TASK_S));
    esCheckError(ES_NPU_GetNumInputTensors(modelId, &numInputTensors));
    if (numInputTensors > 1) {
        ES_LOG_ERR("logits task numInput tensor can not align: %d\n", numInputTensors);
    }

    mLogitsTask.task.inputFdNum = numInputTensors;
    mLogitsTask.task.inputFd[0] = mBlockTasks.back().task.outputFd[OUT_HIDDEN_IDX];
    mLogitsTask.inputMapAddr[0] = mBlockTasks.back().inputMapAddr[OUT_HIDDEN_IDX];

    esCheckError(ES_NPU_GetNumOutputTensors(modelId, &numOutputTensors));
    mLogitsTask.task.outputFdNum = numOutputTensors;
    for (int32_t i = 0; i < numOutputTensors; i++) {
        esCheckError(ES_NPU_GetOutputTensorDesc(modelId, i, &tensor));
        esCheckError(prepareTensorBuffer(tensor, mLogitsTask.task.outputFd[i], &mLogitsTask.outputMapAddr[i], true));
        esCheckError(ES_NPU_Prepare(&mLogitsTask.task.outputFd[i]));
    }

    mLogitsTask.task.modelId = modelId;

    return ES_SUCCESS;
}

ES_U64 Qwen2Client::calcKCacheOutputAddrOffset(int32_t num)
{
    ES_U64 offset;

    return mConfig.precision * mConfig.embeddingDim2 * num;
}

ES_U64 Qwen2Client::calcVCacheOutputAddrOffset(int32_t num)
{
    int32_t dim0 = num / mConfig.tokenN2;
    int32_t dim1 = num % mConfig.tokenN2;

    return mConfig.precision * (mConfig.tokenN2 * mConfig.embeddingDim * dim0 + dim1);
}

void Qwen2Client::runModel(MODEL_TASK_S& modelTask)
{
    // Submit model task
    esCheckError(ES_NPU_Submit(&modelTask.task, 1));
}

int32_t Qwen2Client::argmax(int16_t *data, int32_t size)
{
    int32_t pos = 0;

    for (int32_t i = 1; i < size; i++) {
        if (data[i] > data[pos]) {
            pos = i;
        }
    }

    return pos;
}

prob_data_t Qwen2Client::getTopK(int32_t wordPosInVocab, const int32_t& maskVal, const std::vector<int32_t>& outputIds)
{
    prob_data_t topK;

    auto startTime = std::chrono::high_resolution_clock::now();
    // Get embeddings for next word
    int32_t wordStart = wordPosInVocab * mConfig.embeddingSize;
    int32_t wordEnd = wordStart + mConfig.embeddingSize;
    std::vector<fp16_t> wordEmbedding(mWordEmbeddings + wordStart, mWordEmbeddings + wordEnd);

    // prepare input token
    memcpy(mBlockTasks.front().inputMapAddr[IN_TOKEN_IDX],
           wordEmbedding.data(), wordEmbedding.size() * sizeof(fp16_t));

    if (mConfig.dataDump && checkInterval(maskVal, mConfig.dataTokenInterval)) {
        std::string fileName = "token_" + std::to_string(maskVal) + std::string("_wordembedding.data");
        dumpData(mBlockTasks.front().inputMapAddr[IN_TOKEN_IDX], mBlockTasks.front().task.inputFd[IN_TOKEN_IDX].size, fileName);
    }

    auto modelStartTime = std::chrono::high_resolution_clock::now();

    // run block models
    for (int32_t modelIdx = 0; modelIdx < mBlockTasks.size(); modelIdx++) {
        // adjust output KVCache offset
        mBlockTasks[modelIdx].task.outputFd[OUT_K_CACHE_IDX].offset = calcKCacheOutputAddrOffset(maskVal);
        mBlockTasks[modelIdx].task.outputFd[OUT_V_CACHE_IDX].offset = calcVCacheOutputAddrOffset(maskVal);
        // fill mask
        *static_cast<int16_t*>(mBlockTasks[modelIdx].inputMapAddr[IN_MASK_IDX]) = maskVal;

        ES_LOG_DBG("start runmodel mask=%d, modelIdx=%d, modelId: %d\n", maskVal, modelIdx, mBlockTasks[modelIdx].task.modelId);

        if (mConfig.opDump &&
            checkInterval(maskVal, mConfig.opTokenInterval) &&
            checkInterval(modelIdx, mConfig.opLayerInterval)) {
            NPU_DUMP_CONF_S dumpConf;
            std::string dumpPath = "/tmp/" + std::to_string(maskVal) + "_" + std::to_string(modelIdx);
            char dumpStr[64] = {0};

            snprintf(dumpStr, 64, "%d-%d", mConfig.opListInterval.min, mConfig.opListInterval.max);
            dumpConf.subModelId = mConfig.opType;
            dumpConf.dumpList = dumpStr;

            esCheckError(ES_NPU_ConfigDumpInfo(mBlockTasks[modelIdx].task.modelId, dumpPath.c_str(), &dumpConf, 1));
            esCheckError(ES_NPU_EnableDump(mBlockTasks[modelIdx].task.modelId));
        }

        // run block model
        runModel(mBlockTasks[modelIdx]);

        if (mConfig.opDump &&
            checkInterval(maskVal, mConfig.opTokenInterval) &&
            checkInterval(modelIdx, mConfig.opLayerInterval)) {
            esCheckError(ES_NPU_DisableDump(mBlockTasks[modelIdx].task.modelId));
        }

        if (mConfig.dataDump && checkInterval(maskVal, mConfig.dataTokenInterval)) {
            std::string fileName = DUMP_TOKEN_PREFIX(maskVal, modelIdx) + std::string("_token.data");
            dumpData(mBlockTasks[modelIdx].outputMapAddr[OUT_HIDDEN_IDX],
                     mBlockTasks[modelIdx].task.outputFd[OUT_HIDDEN_IDX].size, fileName);
            fileName = DUMP_TOKEN_PREFIX(maskVal, modelIdx) + std::string("_kcache.data");
            dumpData(mBlockTasks[modelIdx].outputMapAddr[OUT_K_CACHE_IDX],
                     mBlockTasks[modelIdx].task.outputFd[OUT_K_CACHE_IDX].size, fileName);
            fileName = DUMP_TOKEN_PREFIX(maskVal, modelIdx) + std::string("_vcache.data");
            dumpData(mBlockTasks[modelIdx].outputMapAddr[OUT_V_CACHE_IDX],
                     mBlockTasks[modelIdx].task.outputFd[OUT_V_CACHE_IDX].size, fileName);
        }
    }

    auto lmStartTime = std::chrono::high_resolution_clock::now();

    // run logits model
    if (mConfig.logits) {
        runModel(mLogitsTask);
        if (mConfig.dataDump && checkInterval(maskVal, mConfig.dataTokenInterval)) {
            std::string fileName = "token_" + std::to_string(maskVal) + std::string("_logits_out.data");
            dumpData(mLogitsTask.outputMapAddr[0], mLogitsTask.task.outputFd[0].size, fileName);
        }
    }

    auto postStartTime = std::chrono::high_resolution_clock::now();

    // postprocess
    if (mConfig.repetitionPenalty != 1.0) {
        // logits preprocess
        std::unordered_set<int> uniqueOutputIds(outputIds.begin(), outputIds.end());
        uint16_t *data = (uint16_t*)mLogitsTask.outputMapAddr[0];

        for (auto ids : uniqueOutputIds) {
            float tmp = half_float::detail::half2float(data[ids]);
            if (tmp > 0) {
                tmp /= mConfig.repetitionPenalty;
            } else {
                tmp *= mConfig.repetitionPenalty;
            }
            data[ids] = half_float::detail::float2half<std::round_to_nearest>(tmp);
        }
    }

    topK = argmax((int16_t*)mLogitsTask.outputMapAddr[0], mConfig.vocabSize);

    esCheckError(ES_SYS_MemFlushCache(mLogitsTask.task.outputFd[0].memFd));

    auto endTime = std::chrono::high_resolution_clock::now();

    if (mConfig.perfEnable) {
        double preprocessTime = std::chrono::duration_cast<std::chrono::microseconds>(modelStartTime - startTime).count() / 1000.0;
        double modelTime = std::chrono::duration_cast<std::chrono::microseconds>(lmStartTime - modelStartTime).count() / 1000.0;
        double lmTime = std::chrono::duration_cast<std::chrono::microseconds>(postStartTime - lmStartTime).count() / 1000.0;
        double postprocessTime = std::chrono::duration_cast<std::chrono::microseconds>(endTime - postStartTime).count() / 1000.0;
        double tokenTime = std::chrono::duration_cast<std::chrono::microseconds>(endTime - startTime).count() / 1000.0;
        printf("token: %d, preprocess: %fms, model: %fms, lm: %fms postprocess: %fms, time: %fms\n",
               maskVal, preprocessTime, modelTime, lmTime, postprocessTime, tokenTime);
    }

    ES_LOG_MSG("mask=%d topK=%hu\n", maskVal, topK);

    // Return topK value
    return topK;
}

std::vector<std::string> Qwen2Client::splitUtf8(const std::string& str)
{
    std::vector<std::string> result;
    std::string charBuffer;

    const size_t UTF8_1BYTE = 0x80;
    const size_t UTF8_2BYTES_LEAD = 0x6;
    const size_t UTF8_3BYTES_LEAD = 0xe;
    const size_t UTF8_4BYTES_LEAD = 0x1e;

    for (std::size_t i = 0; i < str.size();) {
        charBuffer.clear();
        unsigned char lead = str[i];
        size_t charLen = 0;

        if (lead < UTF8_1BYTE) {
            charLen = 1;
        } else if ((lead >> 5) == UTF8_2BYTES_LEAD) {
            charLen = 2;
        } else if ((lead >> 4) == UTF8_3BYTES_LEAD) {
            charLen = 3;
        } else if ((lead >> 3) == UTF8_4BYTES_LEAD) {
            charLen = 4;
        } else {
            ++i;
            continue;
        }

        charBuffer = str.substr(i, charLen);
        result.push_back(charBuffer);
        ES_LOG_MSG("split word:%s\n", charBuffer.c_str());
        i += charLen;
    }

    return result;
}

std::vector<std::string> Qwen2Client::splitEn(const std::string& input)
{
    std::vector<std::string> tokens;
    std::stringstream ss(input);
    std::string token;
    while (ss >> token) {
        std::string word;
        std::string punctuation;
        for (char c : token) {
            if (ispunct(static_cast<unsigned char>(c))) {
                if (!word.empty()) {
                    tokens.push_back(word);
                    word.clear();
                }
                punctuation += c;
                tokens.push_back(punctuation);
                punctuation.clear();
            } else {
                word += c;
            }
        }
        if (!word.empty()) {
            tokens.push_back(word);
        }
    }

    for (auto& wordStr : tokens) {
        ES_LOG_MSG("wordStr=%s\n", wordStr.c_str());
    }

    return tokens;
}

std::vector<std::string> Qwen2Client::splitSentence(const std::string& input)
{
    if (mConfig.language == "zh") {
        return splitUtf8(input);
    } else {
        return splitEn(input);
    }
}

fp16_t Qwen2Client::fp16Add(fp16_t a, fp16_t b)
{
    float aFloat = half_float::detail::half2float(a);
    float bFloat = half_float::detail::half2float(b);
    float sumFloat = aFloat + bFloat;

    fp16_t result = half_float::detail::float2half<std::round_to_nearest>(sumFloat);

    return result;
}

bool Qwen2Client::checkDuplicateWords(const std::string& sentence)
{
    std::vector<std::string> words = splitSentence(sentence);
    if (words.size() < mConfig.repeatLen * 2) {
        return false;
    }

    std::unordered_map<std::string, int32_t> sequenceCounts;

    for (size_t i = 0; i <= words.size() - mConfig.repeatLen; ++i) {
        std::string subString;
        for (size_t j = i; j < i + mConfig.repeatLen; ++j) {
            subString += words[j];
        }
        sequenceCounts[subString]++;

        if (sequenceCounts[subString] == mConfig.repeatCnt) {
            ES_LOG_MSG("\nduplicate str:%s\n", subString.c_str());
            return true;
        }
    }

    return false;
}

inline bool Qwen2Client::checkInterval(int32_t value, const CLIENT_INTERVAL_S &interval)
{
    return (value >= interval.min && value <= interval.max);
}

std::string Qwen2Client::generate(const std::vector<ChatMessage> &messages)
{
    std::string newSentence = "";
    std::vector<int32_t> sentencePosListInVocab = mTokenizer->encode_messages(messages, mConfig.responseMaxLen);
    std::vector<int32_t> outputIds(sentencePosListInVocab);
    int32_t wordPosInVocab = sentencePosListInVocab[0];
    std::string inputWord;
    int32_t maskVal = 0;
    int32_t lineLength = 0;
    int32_t i;

    ES_LOG_MSG("sentencePosListInVocab size=%u\n", sentencePosListInVocab.size());
    if (sentencePosListInVocab.size() == 0) {
        goto gen_done;
    }

    for (i = 0; i < mConfig.responseMaxLen; ++i) {
        // Skip the original input sentence and print response one by one.
        inputWord = mTokenizer->decode(std::vector<int32_t>(1, wordPosInVocab));

        // stop to generate
        if (i >= sentencePosListInVocab.size()) {
            for (const auto& endStr : mConfig.eosList) {
                if (endStr == inputWord) {
                    ES_LOG_MSG("遇到结束词%s, 生成过程结束。", endStr.c_str());
                    goto gen_done;
                }
            }
        }

        // print generated word.
        if (i >= sentencePosListInVocab.size()) {
            outputIds.emplace_back(wordPosInVocab);
            std::cout << inputWord << std::flush;
        }

        if (lineLength >= mConfig.responseLineLen) {
            if (mConfig.language == "zh") {
                std::cout << std::endl;
            }
            lineLength = 0;
        }
        lineLength += inputWord.size();
        newSentence += inputWord;

        // if (checkDuplicateWords(newSentence)) {
        //     goto gen_done;
        // }
        ES_LOG_MSG("Input sentence_pos(%d) vocab_pos(%d): %s\n", i, wordPosInVocab, inputWord.c_str());

        // Get next word token with top_k value, skip the original input sentence.
        wordPosInVocab = getTopK(wordPosInVocab, maskVal, outputIds);

        if (i < sentencePosListInVocab.size() - 1) {
            wordPosInVocab = sentencePosListInVocab[i + 1];
        }
        ES_LOG_MSG("Output mask(%d) sentence_pos(%d) vocab_pos(%d)\n", maskVal, i + 1, wordPosInVocab);

        maskVal++;
    }

gen_done:
    std::cout << std::endl;

    return newSentence;
}

std::map<int32_t, std::string> Qwen2Client::getOptionalPrompts()
{
    std::map<int32_t, std::string> randomMap;

    if (mConfig.randomCnt >= mConfig.prompts.size()) {
        return mConfig.prompts;
    }

    std::vector<std::map<int32_t, std::string>::const_iterator> iters;
    for (auto it = mConfig.prompts.begin(); it != mConfig.prompts.end(); ++it) {
        iters.push_back(it);
    }

    std::random_device rd;
    std::mt19937 g(rd());
    std::shuffle(iters.begin(), iters.end(), g);

    for (int32_t i = 0; i < mConfig.randomCnt; ++i) {
        randomMap[i] = iters[i]->second;
    }

    return randomMap;
}

std::string Qwen2Client::genResponse(const std::string& prompts)
{
    std::vector<ChatMessage> messages;

    messages.emplace_back(ChatMessage(ChatMessage::ROLE_SYSTEM, "你是一个智能助理."));
    messages.emplace_back(ChatMessage(ChatMessage::ROLE_USER, prompts));

    auto startTime = std::chrono::high_resolution_clock::now();

    std::string response = generate(messages);

    auto endTime = std::chrono::high_resolution_clock::now();
    double duration = std::chrono::duration_cast<std::chrono::microseconds>(endTime - startTime).count();

    if (mConfig.perfEnable) {
        std::cout << "Times:" << (duration / 1000.0 / 1000.0) << "(s)" << std::endl;
    }

    ES_LOG_MSG("response=%s\n", response.c_str());

    return response;
}
